package edu.ntua.ivml.common.lp;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import edu.ntua.ivml.common.utils.Counter;
import edu.ntua.ivml.common.utils.SetUtils;

public class AtomArray {

	public static ArrayList<Atom> create(Atom atom) { 
		ArrayList<Atom> res = new ArrayList<Atom>();
		res.add(atom);
		
		return res;
	}

	public static ArrayList<Atom> apply(ArrayList<Atom> atoms, Substitution subst) {
		ArrayList<Atom> res = new ArrayList<Atom>();
		
		for (Atom a : atoms) {
			Atom newAtom = a.apply(subst);
			if (!res.contains(newAtom)) {
				res.add(newAtom);
			}
		}
		
		return res;
	}
	
	public static Set<Atom> applyS(Set<Atom> atoms, Substitution subst) {
		Set<Atom> res = new HashSet<Atom>();
		
		for (Atom a : atoms) {
			Atom newAtom = a.apply(subst);
			if (!res.contains(newAtom)) {
				res.add(newAtom);
			}
		}
		
		return res;
	}

	public static Set<Variable> getVariables(Collection<Atom> atoms) {
		Set<Variable> res = new HashSet<Variable>();
		
		for (Atom a : atoms) {
			res.addAll(a.getVariables());
		}
		
		return res;
	}

	public static Set<Term> getTerms(Collection<Atom> atoms) {
		Set<Term> res = new HashSet<Term>();
		
		for (Atom a : atoms) {
			res.addAll(a.getTerms());
		}
		
		return res;
	}
	
	public static Map<Variable, Counter> getCountedVariables(ArrayList<Atom> atoms) {
		Map<Variable, Counter> cc = new HashMap<Variable, Counter>();

		Set<Atom> bodyAtoms = new HashSet<Atom>();
			
		for (Atom atom : atoms) {
			if (!bodyAtoms.add(atom)) {
				continue;
			}
				
			for (Term t : atom.getArguments().getTerms()) {
				if (t.isVariable()) {
					Counter c = cc.get((Variable)t);
					if (c == null) {
						c = new Counter(1);
						cc.put((Variable)t, c);
					} else {
						c.increase();
					}
				}
			}
		}
			
		return cc;

	}	

    public static Set<Atom> getAtomsForTerm(Collection<Atom> atoms, Term t) {
    	Set<Atom> res = new HashSet<Atom>(); 
    	for (Atom a : atoms) {
    		if (a.getTerms().contains(t)) {
    			res.add(a);
    		}
    	}
    	
    	return res;
    }
    
    public static int[] termIndexPattern(ArrayList<Atom> atoms, Term t) {
    	int[] res = new int[atoms.size()];
    	Arrays.fill(res, 0);
    	
    	for (int i = 0; i < res.length; i++) {
    		Term[] terms = atoms.get(i).getArguments().getTerms();
    		
    		for (int j = 0; j < terms.length; j++) {
    			if (terms[j].equals(t)) {
    				res[i] += (j + 1);
    			}
    	 	}
    	}
    	
    	return res;
    }
    
	
	public static Subsumee exactSubsumes(ArrayList<Atom> r1, ArrayList<Atom> r2, Set<Term> bound) {
		Set<Atom> body = new HashSet<Atom>();
		body.addAll(r1);
		body.addAll(r2);
		
		Atom head = new Atom(new Predicate("Q", bound.size()), bound.toArray(new Term[] {}));
		Clause clause = new Clause(head, body).condenseWithOriginalVariables();
		
		ArrayList<Atom> cBody = clause.getBody();

		Set<Atom> v1 = SetUtils.intersection(cBody, r1);
		Set<Atom> v2 = SetUtils.intersection(cBody, r2);

		boolean s1 = v1.containsAll(v2);
		boolean s2 = v2.containsAll(v1);
		
		if (!s1 && !s2) {
			if (cBody.size() == r1.size() + r2.size()) {
				return null;
			} else {
				return new Subsumee(cBody, 5);
			}
		} else if (s1 && !s2) {
			return new Subsumee(r1, 2);
		} else if (s2 && !s1) {
			return new Subsumee(r2, 1);
		} else {
			return new Subsumee(r1, 0);
		}	
	}
	
	public static class Subsumee {
		public ArrayList<Atom> subsumee;
		public int mode;
		
		public Subsumee(ArrayList<Atom> subsumee, int mode) {
			this.subsumee = subsumee;
			this.mode = mode;
		}
	}
	

    public static Map<Predicate, AtomIndexNode> getAtomIndex(ArrayList<Atom> r) {
        Map<Predicate, AtomIndexNode> atomIndex = new HashMap<Predicate, AtomIndexNode>();
        
        for (Atom atom : r) {
        	atomIndex.put(atom.getPredicate(), new AtomIndexNode(atom, atomIndex.get(atom.getPredicate())));
        }
        
        return atomIndex;
    }	
    
    public static class AtomIndexNode {
        public final Atom atom;
        public final AtomIndexNode next;
        public boolean matched;

        public AtomIndexNode(Atom atom, AtomIndexNode next) {
            this.atom = atom;
            this.next = next;
            this.matched = false;
        }
        
        public String toString() {
        	return atom + (next == null ? "": ", " + next.toString());
        }
    }

    public static class PredicateComparator implements Comparator<ArrayList<Atom>> {

		public int compare(ArrayList<Atom> a1, ArrayList<Atom> a2) {
			int s1 = a1.size();
			int s2 = a2.size();
			
			if (s1 == 1 && s2 == 1) {
				return a1.get(0).getPredicate().compareTo(a2.get(0).getPredicate());
			} else {
				if (s1 < s2) {
					return -1;
				} else if (s1 > s2) {
					return 1;
				} else {
					return 0;
				}
			}
		}
    }
    
	public static Set<Variable> getUnboundVariables(ArrayList<Atom> atoms, Set<Variable> bound) {
		Map<Variable, Counter> cc = new HashMap<Variable, Counter>();
			
		Set<Atom> bodyAtoms = new HashSet<Atom>();
			
		for (Atom atom : atoms) {
			if (!bodyAtoms.add(atom)) {
				continue;
			}
				
			for (Variable t : atom.getArguments().getVariables()) {
				if (t.isVariable()) {
					Counter c = cc.get(t);
					if (c == null) {
						c = new Counter(1);
						cc.put(t, c);
					} else {
						c.increase();
					}
				}
			}			
		}

		for (Variable v : bound) {
			cc.remove(v);
		}

		Set<Variable> res = new HashSet<Variable>();
			
		for (Map.Entry<Variable, Counter> entry : cc.entrySet()) {
			if (entry.getValue().getValue() == 1) {
				res.add(entry.getKey());
			}
		}
			
		
		return res;

	}	
}
