package edu.ntua.ivml.common.lp;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;


import edu.ntua.ivml.common.lp.ClauseAtomSelector;
import edu.ntua.ivml.common.utils.Counter;
import edu.ntua.ivml.common.utils.Utils;

public class Clause implements ClauseWrapper, Cloneable, Comparable<Clause> {
	
	protected Atom head;
	protected ArrayList<Atom> body;
	
	private int bodyDepth = -1;

	private String canonical;
	private Map<Predicate, ArrayList<Integer>> predicateIndex;
    private Set<Variable> vars = null;
    private Set<Term> terms = null;
    private Set<Constant> cons = null;
	private Set<Term> bTerms = null;
	private Set<Variable> bVars = null;
	private Set<Variable> uVars = null;

	public Clause(Atom headAtom, Atom bodyAtom) {
		this.head = headAtom;
		
		this.body = new ArrayList<Atom>();
		body.add(bodyAtom);
	}
	
	public Clause(Atom headAtom, Collection<Atom> bodyAtoms) {
		this.head = headAtom;

		this.body = new ArrayList<Atom>();
		body.addAll(bodyAtoms);
	}

	public Set<Predicate> getPredicatesForArgument(Tuple t) {
		Set<Predicate> preds = new HashSet<Predicate>();
		
		for (Atom a : body) {
			if (a.args.equals(t)) {
				preds.add(a.getPredicate());
			}
		}
		
		return preds;
	}
	
	public void reset() {
		canonical = null;
		predicateIndex = null;
	    vars = null;
	    terms = null;
	    cons = null;
		bTerms = null;
		bVars = null;
		uVars = null;
	}
	
	public String getCanonical(){
		if (canonical == null) {
			canonical = computeCanonicalRepresentation();
		}

		return canonical;
	}
	
    private String computeCanonicalRepresentation(){
    	Atom[] sBody = body.toArray(new Atom[] {});
    	
    	Arrays.sort(sBody, new Comparator<Atom>(){
    		public int compare(Atom a1, Atom a2){
    			return a1.getFunctionalPrefix().compareTo(a2.getFunctionalPrefix());
    		}
        });

    	String res = head.toString();
        
        if(sBody.length > 0){
            res += " <- ";
             for (int i = 0; i < sBody.length; i++) {
            	 if (i != 0) {
            		res += ", ";
            	 }
            	 res += sBody[i].toString();
             }
        }
        
        return res;
    }
    
	public Object clone() {
		try {
			Clause c = (Clause)super.clone();
			c.head = (Atom)head.clone();
			c.body = new ArrayList<Atom>();
			for (int i = 0; i < body.size(); i++) {
				c.body.add((Atom)body.get(i).clone());
			}
			
			return c;
		} catch (CloneNotSupportedException e) {
			throw new InternalError();
		}				
	}		
	
	private Map<Predicate, ArrayList<Integer>> getPredicateIndex() {
		if (predicateIndex == null) {
			predicateIndex = computePredicateIndex();
		}
		
		return predicateIndex;
	}
	
	private HashMap<Predicate, ArrayList<Integer>> computePredicateIndex() {
		
		HashMap<Predicate, ArrayList<Integer>> index = new HashMap<Predicate, ArrayList<Integer>>();
		
		for (int i = 0; i < body.size(); i++){
			Predicate p = body.get(i).getPredicate();
		
			ArrayList<Integer> list = index.get(p);
			if (list == null) {
				list = new ArrayList<Integer>();
				index.put(p, list);
			}
			list.add(i);
		}

		return index;
	}
	
	public Clause getClause() {
		return this;
	}
	
	public Atom getHead() {
		return head;
	}
	
	public int bodySize() {
		return body.size();
	}

	public ArrayList<Atom> getBody() {
		return body;
	}
	
	public Set<Atom> getBodyAtoms() {
		Set<Atom> res = new HashSet<Atom>();
		res.addAll(body);
		return res;
	}
	
	public Atom getBodyAtomAt(int i) {
		return body.get(i);
	}

	public boolean isBound(Variable v) {
		return getBoundVariables().contains(v);
	}
	
	public boolean isGeneralConcept(Atom atom) {
		return atom.isConcept() ||
	          (atom.isRole() && atom.getArguments().hasSomeTermIn(getUnboundVariables()));
	}
	
	public int bodyDepth() {
		if (bodyDepth == -1) {
			int maxDepth = 0;
		
			for (int i = 0; i < body.size(); i++){
				maxDepth = Math.max(maxDepth, body.get(i).depth());
			}
			
			bodyDepth = maxDepth;
		}
		
		return bodyDepth;
	}
	
	public boolean isTautology() {
        for (int i = 0; i < body.size(); i++) {
        	if (body.get(i).equals(head)) {
        		return true;
        	}
        }
        return false;
    }
	
    public boolean isFunctionFree() {
    	if (head.depth() == 0 && bodyDepth() == 0){
    		return true;
    	}
    	
    	return false;
    }
    
    public int[] termIndexPattern(Term t) {
    	return AtomArray.termIndexPattern(body, t);
    }
    
    public Set<Term> getReachableTerms(Term t) {
    	Set<Term> res = new HashSet<Term>();
    	
    	ArrayList<Term> queue = new ArrayList<Term>();
    	queue.add(t);
    	
    	while (!queue.isEmpty()) {
    		Term ct = queue.remove(0);
    		res.add(ct);
    		
    		for (Term nt : AtomArray.getTerms(AtomArray.getAtomsForTerm(body, ct))) {
    			if (!res.contains(nt)) {
    				queue.add(nt);
    			}
    		}
    	}
    	
    	res.remove(t);
    	
    	return res;
    }
    
    private Set<Variable> getVariableSet() {
		if (vars == null) {
			Set<Variable> res = new HashSet<Variable>();
			res.addAll(head.getVariables());
			
			for (Atom b : body) {
				res.addAll(b.getVariables());
			}
			
			vars = res;
		}
		
		return vars;
    }
    
	public Set<Variable> getVariables() {
		getVariableSet();
		
		return (Set<Variable>)((HashSet<Variable>)vars).clone();
	}
	
	public Set<Term> getTerms() {
		if (terms == null) {
			Set<Term> res = new HashSet<Term>();
			res.addAll(head.getTerms());
			
			for (Atom b : body) {
				res.addAll(b.getTerms());
			}
			
			terms = res;
		} 
		
		return (Set<Term>)((HashSet<Term>)terms).clone();
	}	
	
	public Set<Constant> getConstants() {
		if (cons == null) {
			Set<Constant> res = new HashSet<Constant>();
			for (Term t : head.getTerms()) {
				if (t.isConstant()) {
					res.add((Constant)t);
				}
			}
			
			for (Atom b : body) {
				for (Term t : b.getTerms()) {
					if (t.isConstant()) {
						res.add((Constant)t);
					}
				}
			}
			
			cons = res;
		} 
		
		return (Set<Constant>)((HashSet<Constant>)cons).clone();
	}	
	
	public Set<Predicate> getPredicatesInBody() {
		Set<Predicate> res = new HashSet<Predicate>();
		
		for (Atom a : body) {
			res.add(a.getPredicate());
		}
		
		return res;
	}
	
	
	public Clause apply(Substitution subst) {
		
		ArrayList<Atom> newBody = new ArrayList<Atom>();

		for (int i = 0; i < body.size(); i++) {
			Atom newAtom = body.get(i).apply(subst);
			if (!newBody.contains(newAtom)) {
				newBody.add(newAtom);
			}
		}
		
		
		return new Clause(head.apply(subst), newBody);
	}

	public Clause applyKeepAll(Substitution subst) {
		
		ArrayList<Atom> newBody = new ArrayList<Atom>();

		for (int i = 0; i < body.size(); i++) {
			Atom newAtom = body.get(i).apply(subst);
			newBody.add(newAtom);
		}
		
		
		return new Clause(head.apply(subst), newBody);
	}
	
	public ArrayList<Clause> generateResolvents(ClauseAtomSelector thisClauseSelector, ArrayList<Clause> clauses, ArrayList<ClauseAtomSelector> clauseSelectorMap, Set<Variable> initVariables) {
		ArrayList<Clause> result = new ArrayList<Clause>();

		if (thisClauseSelector.isSelectedHead()) {
			for (int k = 0; k < clauses.size(); k++) {
	    		Clause c = clauses.get(k);
	    		
	        	ClauseAtomSelector cClauseSelector = clauseSelectorMap.get(k);

        		for (int i = 0; i < c.bodySize(); i++){
        			if (cClauseSelector.isSelectedBodyAt(i) && c.getBodyAtomAt(i).getPredicate().equals(this.getHead().getPredicate())) {
        				Clause r = c.resolveMainPremise(i, this, initVariables);
        				if (r != null){
                            result.add(r);
                        }
        			}
        		}
        	}
    	}
    	
		for (int i = 0; i < this.bodySize(); i++) {
			if (thisClauseSelector.isSelectedBodyAt(i)) {
				for (int k = 0; k < clauses.size(); k++) {
					Clause c = clauses.get(k);
					ClauseAtomSelector cClauseSelector = clauseSelectorMap.get(k);

        			if (cClauseSelector.isSelectedHead() && c.getHead().getPredicate().equals(this.getBodyAtomAt(i).getPredicate())) {
        				Clause r = this.resolveMainPremise(i, c, initVariables);
        				if (r != null){
    						result.add(r);
    					}
        			}
        		}
        	}
        }        
        
        return result;
    }
    
	
	public Clause resolveMainPremise(int index, Clause sidePremise, Set<Variable> initVariables) {
		Clause sidePremiseRenamed = sidePremise.apply(Substitution.createDisjointRenaming(getVariableSet(), sidePremise.getVariables()));
		
	    Substitution unifier = Substitution.mgu(getBodyAtomAt(index), sidePremiseRenamed.getHead());
	    
	    if (unifier == null) {
	        return null;
	    }
	    
	    ArrayList<Atom> newBody = new ArrayList<Atom>();
	    
	    for (int i = 0; i < body.size(); i++) {
	        if (i != index) {
	        	Atom newAtom = body.get(i).apply(unifier);
	        	if (!newBody.contains(newAtom)) {
	        		newBody.add(newAtom);
	        	}
	        } else {
	    	    for (int j = 0; j < sidePremiseRenamed.bodySize(); j++){
	    	    	Atom newAtom = sidePremiseRenamed.getBodyAtomAt(j).apply(unifier);
	    	    	if (!newBody.contains(newAtom)) {
	    	    		newBody.add(newAtom);
	    	    	}
	    	    }
	        }
	    }
	    
	    
	    boolean renameVariables = true;
	    Substitution iSubst = new Substitution();
	    
	    for (Variable v : getVariableSet()) {
	    	Term t = unifier.get(v);
	    	if (t != null) {
	    		if (t.isVariable()) {
	    			if (!initVariables.contains(t) && initVariables.contains(v)) {
	    				iSubst.put((Variable)t, v);
	    			}
	    		} else {
	    			renameVariables = false;
	    		}
	    	}
	    }
	    
	    
	    Clause res = new Clause(head.apply(unifier), newBody);
	    
	    if (renameVariables) {
	    	res = res.apply(iSubst);
	    }

	    return res;
	}       


	public int compareTo(Clause r) {
		int c = head.compareTo(r.getHead());
		if (c != 0) {
			return c;
		} else {
			ArrayList<Atom> cBody = r.getBody();
			for (int i = 0; i < Math.min(cBody.size(), body.size()); i++) {
				int nc = body.get(i).compareTo(cBody.get(i));
				if (nc != 0) {
					return nc;
				}
			}
			
			if (body.size() < cBody.size()) {
				return -1;
			} else if (body.size() > cBody.size()) {
				return 1;
			} else {
				return 0;
			}
		}
	}	
	
	public int hashCode() {
		int h = head.getPredicate().hashCode();

		 
		for (Predicate p: getPredicatesInBody()) {
			h += p.hashCode();
		}

		return h;
	}
	
	private class ClauseSubstitution {
		Clause clause;
		Substitution subst;
		
		private ClauseSubstitution(Clause clause, Substitution subst) {
			this.clause = clause;
			this.subst = subst;
		}
		
		public int hashCode() {
			return clause.hashCode();
		}
		
		public boolean equals(Object obj) {
			if (obj instanceof ClauseSubstitution) {
				return clause.equals(((ClauseSubstitution)obj).clause);
			} else {
				return false;
			}
		}
	}
	
	public Clause condense() {

		ArrayList<Clause> unprocessed = new ArrayList<Clause>();
		ArrayList<Clause> condensations = new ArrayList<Clause>();
	
		unprocessed.add(this);
		
		while(!unprocessed.isEmpty()){
			Clause givenClause = unprocessed.remove(0);
			condensations.add(givenClause);

			int givenClauseSize = givenClause.bodySize();
			
			for(int i = 0; i < givenClauseSize - 1; i++){
				Atom givenClauseAtomI = givenClause.getBodyAtomAt(i);
				loop:
				for(int j = i+1; j < givenClauseSize; j++){
					Substitution unifier = Substitution.mgu(givenClauseAtomI, givenClause.getBodyAtomAt(j));
					if (unifier != null){
						Clause newQuery = givenClause.apply(unifier);
						
						for (Clause c: unprocessed){
							if (c.isRenamingOf(newQuery)){
								continue loop;
							}					
						}
						
						for (Clause c: condensations){
							if (c.isRenamingOf(newQuery)){
								continue loop;
							}
						}
						
						unprocessed.add(newQuery);
	                }
				}
			}
		}
		
		Collections.sort(condensations, new Comparator<Clause>(){
			public int compare(Clause c1, Clause c2){
			    return (new Integer(c1.bodySize())).compareTo(new Integer(c2.bodySize()));
			}
		});

		for(Clause c: condensations){
			if (c.subsumes(this)){
				return c;
			}
		}

		return this;
	}

	public Clause condenseWithOriginalVariables() {
		
		ArrayList<ClauseSubstitution> unprocessed = new ArrayList<ClauseSubstitution>();
		ArrayList<ClauseSubstitution> condensations = new ArrayList<ClauseSubstitution>();
	
		unprocessed.add(new ClauseSubstitution(this, new Substitution()));
		
		while(!unprocessed.isEmpty()){
			ClauseSubstitution givenClause = unprocessed.remove(0);
			condensations.add(givenClause);
			
			int givenClauseSize = givenClause.clause.bodySize();

			for(int i = 0; i < givenClauseSize - 1; i++){
				Atom givenClauseAtomI = givenClause.clause.getBodyAtomAt(i);
				loop:
				for(int j = i+1; j < givenClauseSize; j++){
					Substitution unifier = Substitution.mgu(givenClauseAtomI, givenClause.clause.getBodyAtomAt(j));
					if (unifier != null){
						Clause newQuery = givenClause.clause.apply(unifier);

						for (ClauseSubstitution c: unprocessed){
							if (c.clause.isRenamingOf(newQuery)){
								continue loop;
							}					
						}
						
						for (ClauseSubstitution c: condensations){
							if (c.clause.isRenamingOf(newQuery)){
								continue loop;
							}
						}
						
						unprocessed.add(new ClauseSubstitution(newQuery, givenClause.subst.compose(unifier)));
	                }
				}
			}
		}
		
		Collections.sort(condensations, new Comparator<ClauseSubstitution>(){
			public int compare(ClauseSubstitution c1, ClauseSubstitution c2){
			    return (new Integer(c1.clause.bodySize())).compareTo(new Integer(c2.clause.bodySize()));
			}
		});

		Set<Variable> unboundVars = this.getUnboundVariables();
		
		for (ClauseSubstitution c: condensations) {
			if (c.clause.subsumes(this)) {
				Substitution iSubst = new Substitution();
				
			    for (Variable v : c.subst.getVariables()) {
			    	if (unboundVars.contains(c.subst.get(v)) && !unboundVars.contains(v)) {
			    		iSubst.put((Variable)c.subst.get(v), v);
			    	}
			    }
			    
			    removeRedundantAtomsSubstitution = c.subst.compose(iSubst);

				return c.clause.apply(iSubst);
			}
		}
		
		return this;
	
	}
	
	public Substitution removeRedundantAtomsSubstitution;

	public Set<Variable> getBoundVariables() {
		if (bVars == null) {
			Set<Variable> res = getVariables();
			res.removeAll(getUnboundVariables());
			bVars = res;
		}
		
		return (Set<Variable>)((HashSet<Variable>)bVars).clone();
	}
	
	public Set<Variable> getUnboundVariables() {
		if (uVars == null) {
			Map<Variable, Counter> cc = new HashMap<Variable, Counter>();
			
			Set<Atom> bodyAtoms = new HashSet<Atom>();
			
			for (Atom atom : body) {
				if (!bodyAtoms.add(atom)) {
					continue;
				}
				
				for (Term t : atom.getArguments().getTerms()) {
					if (t.isVariable()) {
						Counter c = cc.get(t);
						if (c == null) {
							c = new Counter(1);
							cc.put((Variable)t, c);
						} else {
							c.increase();
						}
					}
				}			
			}

			for (Variable v : head.getVariables()) {
				cc.remove(v);
			}

			Set<Variable> res = new HashSet<Variable>();
			
			for (Map.Entry<Variable, Counter> entry : cc.entrySet()) {
				if (entry.getValue().getValue() == 1) {
					res.add(entry.getKey());
				}
			}
			
			uVars = res;
		}
		
		return (Set<Variable>)((HashSet<Variable>)uVars).clone();

	}	

	public Set<Term> getBoundTerms() {
		if (bTerms == null) {
			Set<Term> res = getTerms();
			res.removeAll(getUnboundVariables());
		
			bTerms = res;
		}
		
		return (Set<Term>)((HashSet<Term>)bTerms).clone();
	}
	

	public Map<Term, Set<Atom>> findSinkAtoms() {
		return findSinkAtomsV();
	}
	
	public Map<Term, Set<Atom>> findSinkAtomsV() {
		Set<Term> headVars = head.getTerms();

		Set<Variable> unboundVars = getUnboundVariables();
		
		Map<Term, ArrayList<Atom>[]> index = new HashMap<Term, ArrayList<Atom>[]>();
		
		for (Atom atom : body) {
			Tuple t = atom.getArguments();
			
			if (atom.isConcept()) {
				ArrayList<Atom>[] set = index.get(t.getTerm(0));
				if (set == null) {
					set = new ArrayList[3];
					index.put(t.getTerm(0), set);
				}
				
				if (set[0] == null) {
					set[0] = new ArrayList<Atom>();
				}
				set[0].add(atom);
			
			} else if (atom.isRole()) {
				boolean[] ub = new boolean[2];
				for (int i = 0; i < 2; i++) {
					ub[i] = unboundVars.contains(t.getTerm(i));
				}

				if (ub[0] || ub[1]) {
					Term iterm;
					if (ub[0]) {
						iterm = t.getTerm(1);
					} else {
						iterm = t.getTerm(0);
					}
					
					ArrayList<Atom>[] set = index.get(iterm);
					if (set == null) {
						set = new ArrayList[3];
						index.put(iterm, set);						
					}
					
					if (set[0] == null) {
						set[0] = new ArrayList<Atom>();
					}
					set[0].add(atom);
					
				} else {
					for (int i = 0; i < 2; i++) {
						ArrayList<Atom>[] set = index.get(t.getTerm(i));
						if (set == null) {
							set = new ArrayList[3];
							index.put(t.getTerm(i), set);
						}
						
						if (set[i + 1] == null) {
							set[i + 1] = new ArrayList<Atom>();
						}
						set[i + 1].add(atom);
					}
				}
			}
		}
		
		Map<Term, Set<Atom>> res = new HashMap<Term, Set<Atom>>();
		
		for (Map.Entry<Term, ArrayList<Atom>[]> entry : index.entrySet()) {
			
			if (headVars.contains(entry.getKey())) {
				continue;
			}
			
			ArrayList<Atom>[] v = entry.getValue();
			
			Set<Atom> group = new HashSet<Atom>();

			if (v[1] != null && v[2] != null) {
				continue;
			} else if (v[1] == null && v[2] != null) {
				group.addAll(v[2]);
			} else if (v[1] != null && v[2] == null ) {
				group.addAll(v[1]);
			}
			
			if (v[0] != null) {
				group.addAll(v[0]);
			}

			if (!group.isEmpty()) {
				res.put(entry.getKey(), group);
			}
		}
		
		return res;
	}


	public <U extends ClauseWrapper> boolean isRedundant(Collection<U> cclauses) {
		return isRedundant(cclauses, false);
	}	

	public <U extends ClauseWrapper> boolean isEQRedundant(Collection<U> cclauses) {
		for (ClauseWrapper rr : cclauses) {
			if (rr.getClause().isRenamingOf(this)) {
				return true;
			} 
		}
		
		return false;
	}	

	public <U extends ClauseWrapper> boolean isRedundant(Collection<U> cclauses, boolean oneToOne) {
		for (ClauseWrapper rr : cclauses) {
			if (rr.getClause().subsumes(this, oneToOne)) {
				return true;
			} 
		}
		
		return false;
	}	

	public <U extends ClauseWrapper> boolean isCCSubsumedBy(U u) {
		return u.getClause().subsumes(this);
	}
	
	// assumes clauses have no redundant atoms
	public static <U extends ClauseWrapper> ArrayList<U> getNonRedundantClauses(ArrayList<U> clauses) {
		return getNonRedundantClauses(clauses, false, null, null, null, null);
	}
	
	public static <U extends ClauseWrapper> ArrayList<U> getNonRedundantClauses(ArrayList<U> clauses, ArrayList<Integer> cids) {
		ArrayList<Integer> ids = new ArrayList<>();

		ArrayList<Integer> nrBlockStart = new ArrayList<>();
		ArrayList<Integer> nrBlockStartIds = new ArrayList<>();

		for (int i = 0; i < clauses.size(); i++) {
			int id = cids.get(i);
		
			if (id == -1) {
				ids = null;
			} else {
				if (nrBlockStart.size() == 0) {
					nrBlockStart.add(0);
					nrBlockStartIds.add(id);
				} else {
					if (nrBlockStartIds.get(nrBlockStart.size() - 1) != id) {
						nrBlockStart.add(ids.size());
						nrBlockStartIds.add(id);
					}
				}
				
				ids.add(id);
			}
		}
		
	
		return getNonRedundantClauses(clauses, false, null, ids, nrBlockStart, nrBlockStartIds);
	}
	
	public static <U extends ClauseWrapper> ArrayList<U> getNonRedundantClauses(ArrayList<U> clauses, boolean oneway, ArrayList<Integer> ids, ArrayList<Integer> blocks, ArrayList<Integer> blockIds) {
		return getNonRedundantClauses(clauses, oneway, null, ids, blocks, blockIds);
	}
	
	public static <U extends ClauseWrapper> ArrayList<U> getNonRedundantClauses(ArrayList<U> clauses, boolean oneway) {
		return getNonRedundantClauses(clauses, oneway, null, null, null, null);
	}
	
	public static <U extends ClauseWrapper> ArrayList<U> getNonRedundantClauses(ArrayList<U> clauses, boolean oneway, Counter ct) {
		return getNonRedundantClauses(clauses, oneway, ct, null, null, null);
	}
	
	private static class LengthComparator<U extends ClauseWrapper> implements Comparator<U> {

		public int compare(U o1, U o2) {
			int i1 = o1.getClause().bodySize();
			int i2 = o2.getClause().bodySize();
			
			if (i1 < i2) {
				return -1;
			} else if (i1 > i2) {
				return 1;
			} else {
				return 0;
			}
		}
	}
	
	private static LengthComparator<ClauseWrapper> lc = new LengthComparator<ClauseWrapper>();
	
	private static class CT implements Comparable<CT> {
		public double len;
		public int start;
		public int end;
		
		public CT(double len, int start, int end) {
			this.len = len;
			this.start = start;
			this.end = end;
		}

		public int compareTo(CT obj) {
			if (len < obj.len) {
				return -1;
			} else if (len > obj.len) {
				return 1;
			} else {
				return 0;
			}
		}
	}
	
	public static <U extends ClauseWrapper> void sort(ArrayList<U> clauses, ArrayList<Integer> ids, ArrayList<Integer> blockStart, ArrayList<Integer> blockStartIds) {
		
		if (ids == null) {
			Collections.sort(clauses, lc);
		} else {
			ArrayList<U> ires = new ArrayList<U>();
			CT[] lens = new CT[blockStart.size()];
			
			int size = blockStart.size();
			for (int i = 0; i < size; i++) {
				int start = blockStart.get(i);
				int end;
				if (i == size - 1) {
					end = clauses.size();
				} else {
					end = blockStart.get(i+1);
				}
				ArrayList<U> tmp = new ArrayList<U>();

				int c = 0;
				for (int j = start; j < end; j++) {
					U cc = clauses.get(j);
					tmp.add(cc);
					c += cc.getClause().bodySize();
				}

				Collections.sort(tmp, lc);
				ires.addAll(tmp);
				
				lens[i] = new CT(((double)c)/(end-start), start, end);			
			}

			Arrays.sort(lens);
			
			ArrayList<U> newClauses = new ArrayList<U>();
			ArrayList<Integer> newIds = new ArrayList<Integer>();
			ArrayList<Integer> newBlockStart = new ArrayList<Integer>();
			ArrayList<Integer> newBlockStartIds = new ArrayList<Integer>();
			
			int id = 0;
			for (CT ct : lens) {
				newBlockStart.add(newClauses.size());
				
				for (int j = ct.start; j < ct.end; j++) {
					newClauses.add(ires.get(j));
					newIds.add(id);
				}
				
				newBlockStartIds.add(id++);
			}
			
			clauses.clear();
			clauses.addAll(newClauses);
			ids.clear();
			ids.addAll(newIds);
			blockStart.clear();
			blockStart.addAll(newBlockStart);
			blockStartIds.clear();
			blockStartIds.addAll(newBlockStartIds);
		}
	}
	
	public static <U extends ClauseWrapper> ArrayList<U> getNonRedundantClauses(ArrayList<U> clauses, boolean oneway, Counter ct, ArrayList<Integer> ids, ArrayList<Integer> blockStart, ArrayList<Integer> blockStartIds) {
		sort(clauses, ids, blockStart, blockStartIds);
		
//		int checkCounter = 0;
		
		ArrayList<U> result = new ArrayList<U>();
		result.addAll(clauses);

		ArrayList<Integer> newIds = null;
		ArrayList<Integer> blocks = null;
		ArrayList<Integer> blocksIds = null;
		
		if (ids != null) {
			newIds = new ArrayList<Integer>();
			newIds.addAll(ids);
			
			blocks = new ArrayList<Integer>();
			blocksIds = new ArrayList<Integer>();
			
			blocks.addAll(blockStart);
			blocksIds.addAll(blockStartIds);
		}
		
//		blocks = null;
		
		ClauseWrapper cc;
		int id = -1;
		int bi = -1;

		int prevId = -1;
		
		if (!oneway) {
			loop1:
			for (int i = 0; i < result.size(); ) {
				cc = result.get(i);
				if (newIds != null) {
					id = newIds.get(i);

					if (blocks != null) {
						if (id != prevId) {
							bi = Utils.binarySearch(blocksIds, id);
							prevId = id;
						}
					}
				}

				for (int j = 0; j < i;) {
					if (ids != null && id == newIds.get(j)) {
						if (blocks != null) {
							if (bi + 1 < blocks.size()) {
								j = blocks.get(bi + 1);
							} else {
								break;
							}
						} else {
							j++;
						}

						continue;
					}
	
//					checkCounter++;
					
					if (cc.isCCSubsumedBy(result.get(j))) {
						if (ct != null && i < ct.getValue()) {
							ct.decrease();
						}
						
						result.remove(i);
						
						if (newIds != null) {
							newIds.remove(i);
							
							if (blocks != null) {
								for (int t = bi + 1; t < blocks.size(); t++) {
									blocks.set(t, blocks.get(t) - 1);
								}

								if (bi + 1 < blocks.size() && blocks.get(bi).equals(blocks.get(bi + 1))) {
									blocks.remove(bi);
									blocksIds.remove(bi);
								}
							}
						}
						continue loop1;
					}
					j++;
				}
				i++;
			}
		}
		
		prevId = -1;
		
		loop2:
		for (int i = 0; i < result.size();) {
			cc = result.get(i);
			if (newIds != null) {
				id = newIds.get(i);
				if (blocks != null) {
					if (id != prevId) {
						bi = Utils.binarySearch(blocksIds, id);
						prevId = id;
					}
				}
			}
			
			for (int j = i + 1; j < result.size();) {
				if (ids != null && id == newIds.get(j)) {
					if (blocks != null) {
						if (bi + 1 < blocks.size()) {
							j = blocks.get(bi + 1);
						} else {
							break;
						}
					} else {
						j++;
					}
					continue;
				}

//				checkCounter++;
				
				if (cc.isCCSubsumedBy(result.get(j))) {
					if (ct != null && i < ct.getValue()) {
						ct.decrease();
					}
					result.remove(i);

					if (newIds != null) {
						newIds.remove(i);
						
						if (blocks != null) {
							for (int t = bi + 1; t < blocks.size(); t++) {
								blocks.set(t, blocks.get(t) - 1);
							}

							if (bi + 1 < blocks.size() && blocks.get(bi).equals(blocks.get(bi + 1))) {
								blocks.remove(bi);
								blocksIds.remove(bi);
							}
						}					
					}
					continue loop2;
				}
				j++;
			}
			i++;
		}
		
		return result;
	}	
	
	public boolean isQueryClause() {
		return getHead().getPredicate().isQueryPredicate();
	}
	
	public String toString() {
		String s = "";
		
		for (Atom b : body) {
			if (s.length() != 0) {
				s += ", ";
			}
			s += b.toString();
		}
		
		return head.toString() + " <- " + s;
	}	
	
	public boolean equals(Object obj) {
		return isRenamingOf((Clause)obj); 
	}

	public boolean isRenamingOf(Clause rule) {
		return isRenamingOfS(rule) != null;
	}
	
    public Substitution isRenamingOfS(Clause rule) {
    	
		if (!head.getPredicate().equals(rule.getHead().getPredicate()) || body.size() != rule.bodySize()) {
			return null;
		}

		Set<Predicate> allPreds = new HashSet<Predicate>();
		for (int i = 0; i < this.body.size(); i++){
			allPreds.add(this.body.get(i).getPredicate());
			allPreds.add(rule.body.get(i).getPredicate());
		}
		
		for (Predicate p : allPreds) {
			ArrayList<Integer> l1 = this.getPredicateIndex().get(p);			
			ArrayList<Integer> l2 = rule.getPredicateIndex().get(p);
			
			if (l1 == null || l2 == null || l1.size() != l2.size()) {
				return null;
			}
		}
		
		Substitution subst = Atom.matchAtoms(head, rule.getHead(), new Substitution());
		if (subst == null) {
			return null;
		}

        Substitution res = match(true, body, AtomArray.getAtomIndex(rule.body), subst, 0);
        
        if (res != null && res.isRenaming(getVariables())) {
        	return res;
        } else {
        	return null;
        }
    }	
	
    public boolean subsumes(Clause rule) {
    	return subsumes(rule, false);
    }
    
    public boolean subsumes(Clause rule, boolean oneToOne) {
    	
		if (!head.getPredicate().equals(rule.getHead().getPredicate())) {
			return false;
		}
				
		Set<Predicate> thatPreds = new HashSet<Predicate>();
		int thatBodySize = rule.bodySize();

		for (int i = 0; i < thatBodySize; i++){			
			thatPreds.add(rule.getBodyAtomAt(i).getPredicate());
		}
		
		for (Predicate p : this.getPredicateIndex().keySet()) {
			if (!thatPreds.contains(p)) {
				return false;
			}
		}			
    	
		Substitution subst = Atom.matchAtoms(head, rule.getHead(), new Substitution());
		if (subst == null) {
			return false;
		} 
		
		ArrayList<Atom> ruleHead = new ArrayList<Atom>();
		ruleHead.add(rule.head);

		Substitution s = match(oneToOne, body, AtomArray.getAtomIndex(rule.getBody()), subst, 0);

        return s != null;
    }

    private Substitution match(boolean oneToOne, ArrayList<Atom> subsumingAtoms, Map<Predicate,AtomArray.AtomIndexNode> subsumedAtomsIndex, Substitution substitution, int index) {
    	
        if (index == subsumingAtoms.size()) {
        	return substitution;
        } else {
            Atom subsumingAtom = subsumingAtoms.get(index);

            AtomArray.AtomIndexNode candidates = subsumedAtomsIndex.get(subsumingAtom.getPredicate());
            
            while (candidates != null) {
            	if (oneToOne){
                    if (!candidates.matched) {
                        Substitution matchedSubstitution = Atom.matchAtoms(subsumingAtom, candidates.atom, substitution);
                        if (matchedSubstitution != null) {
                            candidates.matched = true;
                            Substitution newSubst = match(oneToOne, subsumingAtoms, subsumedAtomsIndex, matchedSubstitution, index + 1); 
                            if (newSubst != null){
                                return newSubst;
                            } else {
                            	candidates.matched = false;
                            }
                        }
                    }
                    candidates = candidates.next;
            	} else {
            		Substitution matchedSubstitution = Atom.matchAtoms(subsumingAtom, candidates.atom, substitution);
                    if (matchedSubstitution != null) {
                    	Substitution newSubst = match(oneToOne, subsumingAtoms, subsumedAtomsIndex, matchedSubstitution, index + 1); 
                        if (newSubst != null) {
                            return newSubst;
                        }
                    }
                    candidates = candidates.next;
            	}
            }
            
            return null;
        }
    }
    
    public class RelevantInfo {
    	public Substitution subst;
    	public ArrayList<Integer> indices;
    	
    	public RelevantInfo(Substitution subst, ArrayList<Integer> indices) {
    		this.subst = subst;
    		this.indices = indices;
    	}
    	
    	public String toString() {
    		return subst + ":" + indices;
    	}
    }
    
    
    public ArrayList<RelevantInfo> relevantSubstitutionsTo(Clause that) {
    	return substitutionsTo(that, true);
    }
    	
	private ArrayList<RelevantInfo> substitutionsTo(Clause that, boolean renaming) {
    	
    	ArrayList<RelevantInfo> res = new ArrayList<RelevantInfo>();
    	
		Set<Predicate> thatPreds = new HashSet<Predicate>();
		int thatBodySize = that.bodySize();

		for (int i = 0; i < thatBodySize; i++){		
			thatPreds.add(that.getBodyAtomAt(i).getPredicate());
		}
		
		for (Predicate p : this.getPredicateIndex().keySet()) {
			if (!thatPreds.contains(p)) {
				return res;
			}
		}			
    	
        ArrayList<Substitution> substArray = new ArrayList<Substitution>();
        substArray.add(new Substitution());

        ArrayList<ArrayList<Integer>> usedArray = new ArrayList<ArrayList<Integer>>();
        usedArray.add(new ArrayList<Integer>());

        matchBody(that, that.getPredicateIndex(), substArray, usedArray, 0);

        for (int i = 0; i < substArray.size(); i++) {
        	Substitution subst = substArray.get(i);
	    	if (!renaming || subst.isRenaming(getVariables())) {
	    		subst.compact();
	    		
	    		res.add(new RelevantInfo(subst, usedArray.get(i)));
	    	}
        }
        
        return res;
    }

    private void matchBody(Clause that, Map<Predicate,ArrayList<Integer>> thatPredMap, ArrayList<Substitution> substArray, ArrayList<ArrayList<Integer>> usedArray, int index) {
    	
        if (index == bodySize()) {
        	return;
        } else {
            Atom currentAtom = getBodyAtomAt(index);

            ArrayList<Integer> candidates = thatPredMap.get(currentAtom.getPredicate());

            int k = substArray.size();
            
            boolean added = false;
            
            for (int i = 0; i < k; i++) {
            	Substitution subst = substArray.remove(0);
            	ArrayList<Integer> used = usedArray.remove(0);
            	
	            for (int j : candidates) {
	            	if (!used.contains(j)) {
	            		Substitution newSubst = Atom.matchAtoms(currentAtom, that.getBodyAtomAt(j), subst);
	            		
	            		if (newSubst != null) {
	            			ArrayList<Integer> newUsed = (ArrayList<Integer>)used.clone();
	            			newUsed.add(j);
	            			
	            			usedArray.add(newUsed);
	            			substArray.add(newSubst);
	            			
	            			added = true;
	            		}
	            	}
	            }
            }
            
            if (added) {
            	matchBody(that, thatPredMap, substArray, usedArray, index + 1);
            }
        }
    }

	public Clause sort(boolean cs) {
		ArrayList<Atom> newbody = new ArrayList<>(body);
		Collections.sort(newbody, new Atom.PredicateComparator(cs));
		
		return new Clause(head, newbody);
	}
}


